@@ -227,9 +227,12 @@ void get_C_pair(TaskData *taskData, PolyWord number, unsigned long *pHi, unsigne
 {
     if ( IS_INT(number) )
     {
-        /* It will fit in the low-order word. */
         *pLo = (unsigned long) UNTAGGED(number);
+#if (SIZEOF_VOIDP > SIZEOF_LONG)
+        *pHi = (unsigned long) (UNTAGGED(number) >> sizeof(unsigned long)*8);
+#else
         *pHi = 0;
+#endif
         return;
     }
     if (OBJ_IS_NEGATIVE(GetLengthWord(number))) raise_exception0(taskData, EXC_size);
@@ -397,7 +400,7 @@ static Handle make_canonical(TaskData *taskData, Handle x, int sign)
 }
 
 
-Handle Make_arbitrary_precision(TaskData *taskData, POLYSIGNED val)
+Handle ArbitraryPrecionFromSigned(TaskData *taskData, POLYSIGNED val)
 /* Called from routines in the run-time system to generate an arbitrary
    precision integer from a word value. */
 {
@@ -421,7 +424,7 @@ Handle Make_arbitrary_precision(TaskData *taskData, POLYSIGNED val)
     return y;
 }
 
-Handle Make_unsigned(TaskData *taskData, POLYUNSIGNED uval)
+Handle ArbitraryPrecionFromUnsigned(TaskData *taskData, POLYUNSIGNED uval)
 /* Called from routines in the run-time system to generate an arbitrary
    precision integer from an unsigned value. */
 {
@@ -442,13 +445,82 @@ Handle Make_unsigned(TaskData *taskData, POLYUNSIGNED uval)
     return y;
 }
 
+
+Handle Make_arbitrary_precision(TaskData *taskData, int val)
+{
+    return ArbitraryPrecionFromSigned(taskData, val);
+}
+
+Handle Make_arbitrary_precision(TaskData *taskData, unsigned uval)
+{
+    return ArbitraryPrecionFromUnsigned(taskData, uval);
+}
+
+Handle Make_arbitrary_precision(TaskData *taskData, long val)
+{
+    return ArbitraryPrecionFromSigned(taskData, val);
+}
+
+Handle Make_arbitrary_precision(TaskData *taskData, unsigned long uval)
+{
+    return ArbitraryPrecionFromUnsigned(taskData, uval);
+}
+
+#if SIZEOF_LONG_LONG != 0
+#if SIZEOF_LONG_LONG <= SIZEOF_VOIDP
+Handle Make_arbitrary_precision(TaskData *taskData, long long val)
+{
+    return ArbitraryPrecionFromSigned(taskData, val);
+}
+
+Handle Make_arbitrary_precision(TaskData *taskData, unsigned long long uval)
+{
+    return ArbitraryPrecionFromUnsigned(taskData, uval);
+}
+#else
+Handle Make_arbitrary_precision(TaskData *taskData, long long val)
+{
+    if (val < 0)
+    {
+        Handle pos;
+
+        // Have to handle the most negative long long specially
+        if (val == -val)
+        {
+            val = -(val + 1LL);
+            pos = Make_arb_from_pair(taskData,
+                                     (unsigned)(val >> (sizeof(unsigned) * 8)),
+                                     (unsigned)val);
+            Handle one = Make_arbitrary_precision(taskData, 1);
+            return neg_longc(taskData, add_longc(taskData, pos, one));
+        }
+        val = -val;
+        pos = Make_arb_from_pair(taskData,
+                                 (unsigned)(val >> (sizeof(unsigned) * 8)),
+                                 (unsigned)val);
+        return neg_longc(taskData, pos);
+    }
+    return Make_arb_from_pair(taskData,
+                              (unsigned)(val >> (sizeof(unsigned) * 8)),
+                              (unsigned)val);
+}
+
+Handle Make_arbitrary_precision(TaskData *taskData, unsigned long long uval)
+{
+    return Make_arb_from_pair(taskData,
+                              (unsigned)(uval >> (sizeof(unsigned) * 8)),
+                              (unsigned)uval);
+}
+#endif
+#endif
+
 /* Creates an arbitrary precision number from two words.
    At present this is used for 64-bit quantities. */
 Handle Make_arb_from_pair(TaskData *taskData, unsigned hi, unsigned lo)
 {
     /* If the high word is zero we can use either the tagged short
        form or a single word. */
-    if (hi == 0) return Make_unsigned(taskData, lo);
+    if (hi == 0) return Make_arbitrary_precision(taskData, lo);
     int words = 2;
     // If the high order byte is non-zero add an extra word for the sign.
     if ((hi >> ((sizeof(hi)-1)*8)) != 0) words++;
@@ -480,9 +552,9 @@ Handle Make_arb_from_pair_scaled(TaskData *taskData, unsigned hi, unsigned lo, u
    /* We might be able to compute the number as a 64 bit quantity and
       then convert it but this is probably more portable. It does risk
       overflowing the save vector. */
-    Handle hHi = Make_unsigned(taskData, hi);
-    Handle hLo = Make_unsigned(taskData, lo);
-    Handle hScale = Make_unsigned(taskData, scale);
+    Handle hHi = Make_arbitrary_precision(taskData, hi);
+    Handle hLo = Make_arbitrary_precision(taskData, lo);
+    Handle hScale = Make_arbitrary_precision(taskData, scale);
     return add_longc(taskData, mult_longc(taskData, hHi, hScale), hLo);
 }
 
