@@ -19,10 +19,12 @@
 
 */
 
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_ASSERT_H
@@ -60,6 +62,7 @@
 #include "diagnostics.h"
 #include "memmgr.h"
 #include "processes.h" // For IO_SPACING
+#include "sys.h" // For EXC_Fail
 
 #include "pexport.h"
 
@@ -84,42 +87,88 @@ have to go back over the memory and turn the tombstones back into length
 words.
 */
 
-CopyScan::CopyScan(unsigned h): hierarchy(h)
+GraveYard::~GraveYard()
+{
+    free(graves);
+}
+
+CopyScan::CopyScan(unsigned h/*=0*/): hierarchy(h)
 {
-    ASSERT(gMem.neSpaces == 0);
-    // Set the space sizes to a quarter the space currently in use.  Making
-    // the spaces too large may be a problem if we're very close to the maximum
-    // address space.  Making them too small may increase the cost of linking.
     defaultImmSize = defaultMutSize = 0;
     defaultNoOverSize = 4096; // This can be small.
+    tombs = 0;
+    graveYard = 0;
+}
+
+void CopyScan::initialise(bool isExport/*=true*/)
+{
+    ASSERT(gMem.neSpaces == 0);
+    // Set the space sizes to a proportion of the space currently in use.
+    // Computing these sizes is not obvious because CopyScan is used both
+    // for export and for saved states.  For saved states in particular we
+    // want to use a smaller size because they are retained after we save
+    // the state and if we have many child saved states it's important not
+    // to waste memory.
+    if (hierarchy == 0)
+    {
+        graveYard = new GraveYard[gMem.npSpaces];
+        if (graveYard == 0)
+            throw MemoryException();
+    }
     unsigned i;
     for (i = 0; i < gMem.npSpaces; i++)
     {
-        MemSpace *space = gMem.pSpaces[i];
-        POLYUNSIGNED size = (space->top-space->bottom)/4;
-        if (space->isMutable)
-            defaultMutSize += size;
-        else
-            defaultImmSize += size;
+        PermanentMemSpace *space = gMem.pSpaces[i];
+        if (space->hierarchy >= hierarchy) {
+            // Include this if we're exporting (hierarchy=0) or if we're saving a state
+            // and will include this in the new state.
+            POLYUNSIGNED size = (space->top-space->bottom)/4;
+            if (space->isMutable)
+                defaultMutSize += size;
+            else
+                defaultImmSize += size;
+            if (space->hierarchy == 0 && ! space->isMutable)
+            {
+                // We need a separate area for the tombstones because this is read-only
+                graveYard[tombs].graves = (PolyWord*)calloc(space->spaceSize(), sizeof(PolyWord));
+                if (graveYard[tombs].graves == 0)
+                    throw MemoryException();
+                graveYard[tombs].startAddr = space->bottom;
+                graveYard[tombs].endAddr = space->top;
+                tombs++;
+            }
+        }
     }
     for (i = 0; i < gMem.nlSpaces; i++)
     {
         LocalMemSpace *space = gMem.lSpaces[i];
-        POLYUNSIGNED size = (space->top-space->pointer)/4;
+        POLYUNSIGNED size = space->allocatedSpace();
+        // It looks as though the mutable size generally gets
+        // overestimated while the immutable size is correct.
         if (space->isMutable)
-            defaultMutSize += size;
+            defaultMutSize += size/4;
         else
-            defaultImmSize += size;
+            defaultImmSize += size/2;
+    }
+    if (isExport)
+    {
+        // Minimum 1M words.
+        if (defaultMutSize < 1024*1024) defaultMutSize = 1024*1024;
+        if (defaultImmSize < 1024*1024) defaultImmSize = 1024*1024;
+    }
+    else
+    {
+        // Much smaller minimum sizes for saved states.
+        if (defaultMutSize < 1024) defaultMutSize = 1024;
+        if (defaultImmSize < 4096) defaultImmSize = 4096;
     }
-    if (defaultMutSize < 1024*1024)
-        defaultMutSize = 1024*1024;
-    if (defaultImmSize < 1024*1024)
-        defaultImmSize = 1024*1024;
 }
 
 CopyScan::~CopyScan()
 {
     gMem.DeleteExportSpaces();
+    if (graveYard)
+        delete[](graveYard);
 }
 
 
@@ -176,6 +225,26 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
         *pt = newAddr;
         return 0; // No need to scan it again.
     }
+    else if (space->spaceType == ST_PERMANENT)
+    {
+        // See if we have this in the grave-yard.
+        for (unsigned i = 0; i < tombs; i++)
+        {
+            GraveYard *g = &graveYard[i];
+            if (val.AsStackAddr() >= g->startAddr && val.AsStackAddr() < g->endAddr)
+            {
+                PolyWord *tombAddr = g->graves + (val.AsStackAddr() - g->startAddr);
+                PolyObject *tombObject = (PolyObject*)tombAddr;
+                if (tombObject->ContainsForwardingPtr())
+                {
+                    *pt = tombObject->GetForwardingPtr();;
+                    return 0;
+                }
+                break; // No need to look further
+            }
+        }
+    }
+
     // No, we need to copy it.
     ASSERT(space->spaceType == ST_LOCAL || space->spaceType == ST_PERMANENT);
     POLYUNSIGNED lengthWord = obj->LengthWord();
@@ -184,12 +253,19 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
     PolyObject *newObj = 0;
     bool isMutableObj = obj->IsMutable();
     bool isNoOverwrite = false;
-    if (isMutableObj) isNoOverwrite = obj->IsNoOverwriteObject();
+    bool isByteObj = false;
+    if (isMutableObj)
+    {
+        isNoOverwrite = obj->IsNoOverwriteObject();
+        isByteObj = obj->IsByteObject();
+    }
     // Allocate a new address for the object.
     for (unsigned i = 0; i < gMem.neSpaces; i++)
     {
         PermanentMemSpace *space = gMem.eSpaces[i];
-        if (isMutableObj == space->isMutable && isNoOverwrite == space->noOverwrite)
+        if (isMutableObj == space->isMutable &&
+            isNoOverwrite == space->noOverwrite &&
+            isByteObj == space->byteOnly)
         {
             ASSERT(space->topPointer <= space->top && space->topPointer >= space->bottom);
             POLYUNSIGNED spaceLeft = space->top - space->topPointer;
@@ -211,6 +287,7 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
         if (spaceWords <= words)
             spaceWords = words+1; // Make sure there's space for this object.
         PermanentMemSpace *space = gMem.NewExportSpace(spaceWords, isMutableObj, isNoOverwrite);
+        if (isByteObj) space->byteOnly = true;
         if (space == 0)
         {
             // Unable to allocate this.
@@ -223,12 +300,26 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
 
     newObj->SetLengthWord(lengthWord); // copy length word
 
-    if (OBJ_IS_STACK_OBJECT(lengthWord))
-        CopyStackFrame ((StackObject *)obj,(StackObject *)newObj);
-    else
-        memcpy(newObj, obj, words*sizeof(PolyWord));
+    memcpy(newObj, obj, words*sizeof(PolyWord));
 
-    obj->SetForwardingPtr(newObj); // Put forwarding pointer in old object.
+    if (space->spaceType == ST_PERMANENT && !space->isMutable && ((PermanentMemSpace*)space)->hierarchy == 0)
+    {
+        // The immutable permanent areas are read-only.
+        unsigned m;
+        for (m = 0; m < tombs; m++)
+        {
+            GraveYard *g = &graveYard[m];
+            if (val.AsStackAddr() >= g->startAddr && val.AsStackAddr() < g->endAddr)
+            {
+                PolyWord *tombAddr = g->graves + (val.AsStackAddr() - g->startAddr);
+                PolyObject *tombObject = (PolyObject*)tombAddr;
+                tombObject->SetForwardingPtr(newObj);
+                break; // No need to look further
+            }
+        }
+        ASSERT(m < tombs); // Should be there.
+    }
+    else obj->SetForwardingPtr(newObj); // Put forwarding pointer in old object.
 
     if (OBJ_IS_CODE_OBJECT(lengthWord))
     {
@@ -277,7 +368,10 @@ static POLYUNSIGNED GetObjLength(PolyObject *obj)
             obj->SetLengthWord(length);
         return length;
     }
-    else return obj->LengthWord();
+    else {
+        ASSERT(obj->ContainsNormalLengthWord());
+        return obj->LengthWord();
+    }
 }
 
 static void FixForwarding(PolyWord *pt, POLYUNSIGNED space)
@@ -288,6 +382,7 @@ static void FixForwarding(PolyWord *pt, POLYUNSIGNED space)
         PolyObject *obj = (PolyObject*)pt;
         POLYUNSIGNED length = OBJ_OBJECT_LENGTH(GetObjLength(obj));
         pt += length;
+        ASSERT(space > length);
         space -= length+1;
     }
 }
@@ -295,7 +390,8 @@ static void FixForwarding(PolyWord *pt, POLYUNSIGNED space)
 class ExportRequest: public MainThreadRequest
 {
 public:
-    ExportRequest(Handle root, Exporter *exp): exportRoot(root), exporter(exp) {}
+    ExportRequest(Handle root, Exporter *exp): MainThreadRequest(MTP_EXPORTING),
+        exportRoot(root), exporter(exp) {}
 
     virtual void Perform() { exporter->RunExport(exportRoot->WordP()); }
     Handle exportRoot;
@@ -332,11 +428,13 @@ static void exporter(TaskData *taskData, Handle args, const char *extension, Exp
 void Exporter::RunExport(PolyObject *rootFunction)
 {
     Exporter *exports = this;
-    // Copy the root and everything reachable from it into the temporary area.
-    CopyScan copyScan;
 
     PolyObject *copiedRoot = 0;
+    CopyScan copyScan;
+
     try {
+        copyScan.initialise();
+        // Copy the root and everything reachable from it into the temporary area.
         copiedRoot = copyScan.ScanObjectAddress(rootFunction);
     }
     catch (MemoryException)
@@ -351,7 +449,8 @@ void Exporter::RunExport(PolyObject *rootFunction)
     {
         LocalMemSpace *space = gMem.lSpaces[j];
         // Local areas only have objects from the allocation pointer to the top.
-        FixForwarding(space->pointer, space->top - space->pointer);
+        FixForwarding(space->bottom, space->lowerAllocPtr - space->bottom);
+        FixForwarding(space->upperAllocPtr, space->top - space->upperAllocPtr);
     }
     for (j = 0; j < gMem.npSpaces; j++)
     {
@@ -375,7 +474,7 @@ void Exporter::RunExport(PolyObject *rootFunction)
     MemSpace *ioSpace = gMem.IoSpace();
     exports->memTable[0].mtAddr = ioSpace->bottom;
     exports->memTable[0].mtLength = (char*)ioSpace->top - (char*)ioSpace->bottom;
-    exports->memTable[0].mtFlags = 0;
+    exports->memTable[0].mtFlags = MTF_WRITEABLE; // Needs to be written during initialisation.
     exports->memTable[0].mtIndex = 0;
 
     for (unsigned i = 0; i < gMem.neSpaces; i++)
@@ -392,6 +491,7 @@ void Exporter::RunExport(PolyObject *rootFunction)
         }
         else
             entry->mtFlags = MTF_EXECUTABLE;
+        if (space->byteOnly) entry->mtFlags |= MTF_BYTES;
     }
 
     exports->memTableEntries = gMem.neSpaces+1;
@@ -408,7 +508,7 @@ Handle exportNative(TaskData *taskData, Handle args)
 {
 #ifdef HAVE_PECOFF
     // Windows including Cygwin
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
     const char *extension = ".obj"; // Windows
 #else
     const char *extension = ".o"; // Cygwin
@@ -488,43 +588,6 @@ void Exporter::relocateObject(PolyObject *p)
         for (POLYUNSIGNED i = 0; i < constCount; i++) relocateValue(&(cp[i]));
 
     }
-    else if (p->IsStackObject())
-    {
-        StackObject *s = (StackObject*)p;
-        POLYUNSIGNED length = p->Length();
-
-        ASSERT(! p->IsMutable()); // Should have been frozen
-        /* First the standard registers, space, pc, sp, hr. */
-        // pc may be TAGGED(0) indicating a retry.
-        PolyWord pc = PolyWord::FromCodePtr(s->p_pc);
-        if (pc != TAGGED(0))
-            s->p_pc = createRelocation(pc, &s->p_pc).AsCodePtr();
-
-        PolyWord *stackPtr = s->p_sp; // Save this before we change it.
-        s->p_sp = createRelocation(PolyWord::FromStackAddr(s->p_sp), &s->p_sp).AsStackAddr();
-        s->p_hr = createRelocation(PolyWord::FromStackAddr(s->p_hr), &s->p_hr).AsStackAddr();
-
-        /* Checked registers. */
-        PolyWord *stackStart = (PolyWord*)p;
-        PolyWord *stackEnd = stackStart+length;
-        for (POLYUNSIGNED i = 0; i < s->p_nreg; i++)
-        {
-            PolyWord r = s->p_reg[i];
-            if (r.AsStackAddr() >= stackStart && r.AsStackAddr() < stackEnd)
-                createRelocation(&s->p_reg[i]);
-            /* It seems we can have zeros in the registers, at least on the i386. */
-            else if (r == PolyWord::FromUnsigned(0)) {}
-            else relocateValue(&(s->p_reg[i]));
-        }
-        /* Now the values on the stack. */
-        for (PolyWord *q = stackPtr; q < stackEnd; q++)
-        {
-            /* A stack may contain a value which is an offset. */
-            if ((*q).AsStackAddr() >= stackStart && (*q).AsStackAddr() < stackEnd)
-                createRelocation(q);
-            else relocateValue(q);
-        }
-    }
     else /* Ordinary objects, essentially tuples. */
     {
         POLYUNSIGNED length = p->Length();
@@ -544,7 +607,7 @@ ExportStringTable::~ExportStringTable()
 // Add a string to the string table, growing it if necessary.
 unsigned long ExportStringTable::makeEntry(const char *str)
 {
-    int len = strlen(str);
+    unsigned len = (unsigned)strlen(str);
     unsigned long entry = stringSize;
     if (stringSize + len + 1 > stringAvailable)
     {
