@@ -106,7 +106,7 @@ public:
     virtual void CallIO5(TaskData *taskData, Handle(*ioFun)(TaskData *, Handle, Handle, Handle, Handle, Handle)) = 0;
     // These next two are sufficiently different that they need to be implemented
     // as special cases.
-    virtual void SetExceptionTrace(TaskData *taskData) = 0;
+    virtual void SetExceptionTrace(TaskData *taskData, bool isLegacy) = 0;
     virtual void CallCodeTupled(TaskData *taskData) = 0;
     // This is used to get the argument to the callback_result function.
     virtual Handle CallBackResult(TaskData *taskData) = 0;
@@ -116,10 +116,11 @@ public:
 
     // Increment or decrement the first word of the object pointed to by the
     // mutex argument and return the new value.
-    virtual Handle AtomicIncrement(TaskData *taskData, Handle mutexp);
-    virtual Handle AtomicDecrement(TaskData *taskData, Handle mutexp);
-    // Set a mutex to one.  On the Sparc this requires a memory barrier.
-    virtual void SetToReleased(TaskData *taskData, Handle mutexp);
+    virtual Handle AtomicIncrement(TaskData *taskData, Handle mutexp) = 0;
+    virtual Handle AtomicDecrement(TaskData *taskData, Handle mutexp) = 0;
+    // Reset a mutex to one.  This needs to be atomic with respect to the
+    // atomic increment and decrement instructions.
+    virtual void AtomicReset(TaskData *taskData, Handle mutexp) = 0;
 };
 
 extern MachineDependent *machineDependent;
