@@ -19,10 +19,12 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_STDIO_H
@@ -49,7 +51,7 @@
 #include "machine_dep.h"
 #include "scanaddrs.h"
 #include "run_time.h"
-#include "polyexports.h"
+#include "../polyexports.h"
 #include "version.h"
 #include "sys.h"
 #include "polystring.h"
@@ -108,7 +110,7 @@ void PExport::printCodeAddr(byte *q)
 {
     PolyObject *obj = ObjCodePtrToPtr(q);
     unsigned long a = getIndex(obj);
-    fprintf(exportFile, "$%lu+%lu", a, (POLYUNSIGNED)(q - (byte*)obj));
+    fprintf(exportFile, "$%lu+%" POLYUFMT, a, (POLYUNSIGNED)(q - (byte*)obj));
 }
 
 /* Get the index corresponding to an address. */
@@ -119,8 +121,8 @@ void PExport::printAddress(void *p)
     {
         // Is it an IO entry?
         POLYUNSIGNED byteOffset = (char*)p - (char*)memTable[area].mtAddr;
-        unsigned ioEntry = byteOffset / (ioSpacing*sizeof(PolyWord));
-        unsigned ioOffset = byteOffset - ioEntry * (ioSpacing*sizeof(PolyWord));
+        unsigned ioEntry = (unsigned)(byteOffset / (ioSpacing*sizeof(PolyWord)));
+        unsigned ioOffset = (unsigned)(byteOffset - ioEntry * (ioSpacing*sizeof(PolyWord)));
         ASSERT(ioEntry >= 0 && ioEntry < POLY_SYS_vecsize);
         if (ioOffset == 0)
            fprintf(exportFile, "I%d", ioEntry);
@@ -134,7 +136,7 @@ void PExport::printAddress(void *p)
 void PExport::printValue(PolyWord q)
 {
     if (IS_INT(q) || q == PolyWord::FromUnsigned(0))
-        fprintf(exportFile, "%ld", UNTAGGED(q));
+        fprintf(exportFile, "%" POLYSFMT, UNTAGGED(q));
     else if (OBJ_IS_CODEPTR(q))
         printCodeAddr(q.AsCodePtr());
     else
@@ -169,11 +171,13 @@ void PExport::printObject(PolyObject *p)
            represented by the character. */
         /* This is not infallible but it seems to be good enough
            to detect the strings. */
-        if (ps->length > 1 &&
-            (POLYUNSIGNED)((ps->length + sizeof(PolyWord) -1) / sizeof(PolyWord)) == length-1)
+        POLYUNSIGNED bytes = length * sizeof(PolyWord);
+        if (length >= 2 &&
+            ps->length <= bytes - sizeof(POLYUNSIGNED) &&
+            ps->length > bytes - 2 * sizeof(POLYUNSIGNED))
         {
             /* Looks like a string. */
-            fprintf(exportFile, "S%lu|", ps->length);
+            fprintf(exportFile, "S%" POLYUFMT "|", ps->length);
             for (unsigned i = 0; i < ps->length; i++)
             {
                 char ch = ps->chars[i];
@@ -199,7 +203,7 @@ void PExport::printObject(PolyObject *p)
             */
             byte *u = (byte*)p;
             putc('B', exportFile);
-            fprintf(exportFile, "%lu|", length*sizeof(PolyWord));
+            fprintf(exportFile, "%" POLYUFMT "|", length*sizeof(PolyWord));
             for (unsigned i = 0; i < (unsigned)(length*sizeof(PolyWord)); i++)
             {
                 fprintf(exportFile, "%02x", u[i]);
@@ -215,11 +219,12 @@ void PExport::printObject(PolyObject *p)
            number of constants. */
         p->GetConstSegmentForCode(cp, constCount);
         /* The byte count is the length of the segment minus the
-           number of constant minus one for the constant count, one for the
-           marker word, one for the byte count and one for the
-           profile count. */
-        POLYUNSIGNED byteCount = (length - constCount - 4) * sizeof(PolyWord);
-        fprintf(exportFile, "C%lu,%lu|", constCount, byteCount);
+           number of constants minus one for the constant count.
+           It includes the marker word, byte count, profile count
+           and, on the X86/64 at least, any non-address constants.
+           These are actually word values. */
+        POLYUNSIGNED byteCount = (length - constCount - 1) * sizeof(PolyWord);
+        fprintf(exportFile, "D%" POLYUFMT ",%" POLYUFMT "|", constCount, byteCount);
 
         // First the code.
         byte *u = (byte*)p;
@@ -238,72 +243,9 @@ void PExport::printObject(PolyObject *p)
         // Finally any constants in the code object.
         machineDependent->ScanConstantsWithinCode(p, this);
     }
-    else if (p->IsStackObject())
-    {
-        StackObject *s = (StackObject*)p;
-        PolyWord *q;
-        ASSERT(! p->IsMutable());
-        fprintf(exportFile, "Q%lu|", length);
-        /* First the standard registers, space, pc, sp, hr. */
-        fprintf(exportFile, "%lu,", s->p_space);
-
-        /* pc may be TAGGED(0) indicating a retry. */
-        PolyWord pc = PolyWord::FromCodePtr(s->p_pc);
-        if (pc == TAGGED(0))
-            fprintf(exportFile, "%lu,\n", TAGGED(0).AsUnsigned());
-        else printCodeAddr(s->p_pc);
-
-        putc(',', exportFile);
-        fprintf(exportFile, "%%%lu+%lu,", myIndex, (POLYUNSIGNED)(s->p_sp-(PolyWord*)p)); /* Word offset of sp. */
-        fprintf(exportFile, "%%%lu+%lu", myIndex, (POLYUNSIGNED)(s->p_hr-(PolyWord*)p)); /* Word offset of hr. */
-
-        /* Checked registers. */
-        fprintf(exportFile, " %lu|", s->p_nreg);
-        PolyWord *stackStart = (PolyWord*)p;
-        PolyWord *stackEnd = stackStart+length;
-        for (i = 0; i < s->p_nreg; i++)
-        {
-            PolyWord r = s->p_reg[i];
-            if (r.AsStackAddr() >= stackStart && r.AsStackAddr() < stackEnd)
-                fprintf(exportFile, "%%%lu+%lu", myIndex, (POLYUNSIGNED)(r.AsStackAddr() - (PolyWord*)p));
-            /* It seems we can have zeros in the registers, at least on the i386. */
-            else if (r == PolyWord::FromUnsigned(0))
-                fprintf(exportFile, "0");
-            else
-                printValue(r);
-            if (i < s->p_nreg-1)
-                putc(',', exportFile);
-        }
-        /* Unchecked registers, just as numbers. */
-        POLYUNSIGNED nUnchecked = s->p_reg[i++].AsUnsigned();
-        printf(" %lu|", nUnchecked);
-        nUnchecked += i;
-        for (; i < nUnchecked; i++)
-        {
-            fprintf(exportFile, "%lu", s->p_reg[i].AsUnsigned());
-            if (i < nUnchecked-1)
-                putc(',', exportFile);
-        }
-        q = s->p_sp;
-        /* Now the values on the stack. */
-        POLYUNSIGNED stackLength = length - (s->p_sp-stackStart);
-        fprintf(exportFile, " %lu|", stackLength);
-        q = s->p_sp;
-        for (i = 0; i < stackLength; i++)
-        {
-            PolyWord r = q[i];
-            /* A stack may contain a value which is an offset. */
-            if (r.AsStackAddr() >= stackStart && r.AsStackAddr() < stackEnd)
-                fprintf(exportFile, "%%%lu+%lu", myIndex, (POLYUNSIGNED)(r.AsStackAddr() - (PolyWord*)p));
-            else printValue(r);
-            if (i < stackLength-1)
-                putc(',', exportFile);
-        }
-
-    }
     else /* Ordinary objects, essentially tuples. */
     {
-        fprintf(exportFile, "O%lu|", length);
+        fprintf(exportFile, "O%" POLYUFMT "|", length);
         for (i = 0; i < length; i++)
         {
             printValue(p->Get(i));
@@ -323,7 +265,7 @@ void PExport::ScanConstant(byte *addr, ScanRelocationKind code)
     // We put in all the values including tagged constants.
     PolyObject *obj = ObjCodePtrToPtr(addr);
     // Put in the byte offset and the relocation type code.
-    fprintf(exportFile, "%lu,%d,", (POLYUNSIGNED)(addr - (byte*)obj), code);
+    fprintf(exportFile, "%" POLYUFMT ",%d,", (POLYUNSIGNED)(addr - (byte*)obj), code);
     printValue(p); // The value to plug in.
     fprintf(exportFile, " ");
 }
@@ -341,7 +283,7 @@ void PExport::exportStore(void)
     {
         if (i != ioMemEntry)
         {
-            totalBytes += memTable[i].mtLength;
+            totalBytes += (unsigned long)memTable[i].mtLength;
             // Get the lowest address.
             if (startAddr == 0 || memTable[i].mtAddr < startAddr)
                 startAddr = memTable[i].mtAddr;
@@ -473,7 +415,7 @@ bool SpaceAlloc::AddToTable(void)
     if (base != 0)
     {
         // Add the new space to the permanent memory table.
-        MemSpace* space = gMem.NewPermanentSpace(base, used, isMutable, false, spaceIndex++);
+        MemSpace* space = gMem.NewPermanentSpace(base, used, isMutable ? MTF_WRITEABLE : 0, spaceIndex++);
         if (space == 0)
         {
             fprintf(stderr, "Insufficient memory\n");
@@ -560,7 +502,7 @@ bool PImport::GetValue(PolyWord *result)
     {
         /* Address of an object. */
         POLYUNSIGNED obj;
-        fscanf(f, "%lu", &obj);
+        fscanf(f, "%" POLYUFMT, &obj);
         ASSERT(obj < nObjects);
         *result = objMap[obj];
     }
@@ -568,7 +510,7 @@ bool PImport::GetValue(PolyWord *result)
     {
         /* Code address. */
         POLYUNSIGNED obj, offset;
-        fscanf(f, "%lu+%lu", &obj, &offset);
+        fscanf(f, "%" POLYUFMT "+%" POLYUFMT, &obj, &offset);
         ASSERT(obj < nObjects);
         PolyObject *q = objMap[obj];
         ASSERT(q->IsCodeObject());
@@ -579,38 +521,27 @@ bool PImport::GetValue(PolyWord *result)
         /* Tagged integer. */
         POLYSIGNED j;
         ungetc(ch, f);
-        fscanf(f, "%ld", &j);
+        fscanf(f, "%" POLYSFMT, &j);
         /* The assertion may be false if we are porting to a machine
            with a shorter tagged representation. */
         ASSERT(j >= -MAXTAGGED-1 && j <= MAXTAGGED);
         *result = TAGGED(j);
     }
-    else if (ch == '%')
-    {
-        /* Offset within the object.  Only in a stack. */
-        POLYUNSIGNED obj, offset;
-        fscanf(f, "%lu+%lu", &obj, &offset);
-        ASSERT(obj < nObjects);
-        PolyObject *q = objMap[obj];
-        ASSERT(q->IsStackObject());
-        ASSERT(offset >= 0 && offset < q->Length());
-        *result = PolyWord::FromStackAddr(((PolyWord)q).AsStackAddr() + offset);
-    }
     else if (ch == 'I')
     {
         /* IO entry number. */
         POLYUNSIGNED j;
-        fscanf(f, "%lu", &j);
+        fscanf(f, "%" POLYUFMT, &j);
         ASSERT(j >= 0 && j < POLY_SYS_vecsize);
-        *result = (PolyObject*)&gMem.ioSpace.bottom[j * IO_SPACING];
+        *result = (PolyObject*)&gMem.ioSpace->bottom[j * IO_SPACING];
     }
     else if (ch == 'J')
     {
         /* IO entry number with offset. */
         POLYUNSIGNED j, offset;
-        fscanf(f, "%lu+%lu", &j, &offset);
+        fscanf(f, "%" POLYUFMT "+%" POLYUFMT, &j, &offset);
         ASSERT(j >= 0 && j < POLY_SYS_vecsize);
-        PolyWord base = (PolyObject*)&gMem.ioSpace.bottom[j * IO_SPACING];
+        PolyWord base = (PolyObject*)&gMem.ioSpace->bottom[j * IO_SPACING];
         *result = PolyWord::FromCodePtr(base.AsCodePtr() + offset);
     }
     else
@@ -640,7 +571,7 @@ bool PImport::DoImport()
 
     ASSERT(gMem.npSpaces == 0);
     ASSERT(gMem.neSpaces == 0);
-    ASSERT(gMem.ioSpace.bottom == 0);
+    ASSERT(gMem.ioSpace->bottom == 0);
     PolyWord *ioSpace = (PolyWord*)calloc(POLY_SYS_vecsize*IO_SPACING, sizeof(PolyWord));
     if (ioSpace == 0)
     {
@@ -654,7 +585,7 @@ bool PImport::DoImport()
     if (ch == 'M') { while (getc(f) != '\n') ; ch = getc(f); }
     ASSERT(ch == 'O'); /* Number of objects. */
     while (getc(f) != '\t') ;
-    fscanf(f, "%lu", &nObjects);
+    fscanf(f, "%" POLYUFMT, &nObjects);
     /* Create a mapping table. */
     objMap = (PolyObject**)calloc(nObjects, sizeof(PolyObject*));
     if (objMap == 0)
@@ -669,7 +600,7 @@ bool PImport::DoImport()
     } while (ch == '\n');
     ASSERT(ch == 'R'); /* Root object number. */
     while (getc(f) != '\t') ;
-    fscanf(f, "%lu", &nRoot);
+    fscanf(f, "%" POLYUFMT, &nRoot);
 
     /* Now the objects themselves. */
     while (1)
@@ -683,7 +614,7 @@ bool PImport::DoImport()
         } while (ch == '\r' || ch == '\n');
         if (ch == EOF) break;
         ungetc(ch, f);
-        fscanf(f, "%lu", &objNo);
+        fscanf(f, "%" POLYUFMT, &objNo);
         ch = getc(f);
         ASSERT(ch == ':');
         ASSERT(objNo < nObjects);
@@ -701,15 +632,13 @@ bool PImport::DoImport()
         /* Object type. */
         switch (ch)
         {
-        case 'Q': /* Stack segment. */
-            objBits |= F_STACK_OBJ;
         case 'O': /* Simple object. */
-            fscanf(f, "%lu", &nWords);
+            fscanf(f, "%" POLYUFMT, &nWords);
             break;
 
         case 'B': /* Byte segment. */
             objBits |= F_BYTE_OBJ;
-            fscanf(f, "%lu", &nBytes);
+            fscanf(f, "%" POLYUFMT, &nBytes);
             /* Round up to appropriate number of words. */
             nWords = (nBytes + sizeof(PolyWord) -1) / sizeof(PolyWord);
             break;
@@ -717,18 +646,19 @@ bool PImport::DoImport()
         case 'S': /* String. */
             objBits |= F_BYTE_OBJ;
             /* The length is the number of characters. */
-            fscanf(f, "%lu", &nBytes);
+            fscanf(f, "%" POLYUFMT, &nBytes);
             /* Round up to appropriate number of words.  Need to add
                one PolyWord for the length PolyWord.  */
             nWords = (nBytes + sizeof(PolyWord) -1) / sizeof(PolyWord) + 1;
             break;
 
-        case 'C': /* Code segment. */
+        case 'C': /* Code segment (old form). */
+        case 'D': /* Code segment (new form). */
             objBits |= F_CODE_OBJ;
             /* Read the number of bytes of code and the number of words
                for constants. */
-            fscanf(f, "%lu,%lu", &nWords, &nBytes);
-            nWords += 4; /* Add words for extras. */
+            fscanf(f, "%" POLYUFMT ",%" POLYUFMT, &nWords, &nBytes);
+            nWords += ch == 'C' ? 4 : 1; /* Add words for extras. */
             /* Add in the size of the code itself. */
             nWords += (nBytes + sizeof(PolyWord) -1) / sizeof(PolyWord);
             break;
@@ -769,7 +699,7 @@ bool PImport::DoImport()
         POLYUNSIGNED  nWords, nBytes, i;
         if (feof(f))
             break;
-        fscanf(f, "%lu", &objNo);
+        fscanf(f, "%" POLYUFMT, &objNo);
         if (feof(f))
             break;
         ch = getc(f);
@@ -787,7 +717,7 @@ bool PImport::DoImport()
         switch (ch)
         {
         case 'O': /* Simple object. */
-            fscanf(f, "%lu", &nWords);
+            fscanf(f, "%" POLYUFMT, &nWords);
             ch = getc(f);
             ASSERT(ch == '|');
             ASSERT(nWords == p->Length());
@@ -806,7 +736,7 @@ bool PImport::DoImport()
         case 'B': /* Byte segment. */
             {
                 byte *u = (byte*)p;
-                fscanf(f, "%lu", &nBytes);
+                fscanf(f, "%" POLYUFMT, &nBytes);
                 ch = getc(f); ASSERT(ch == '|');
                 for (i = 0; i < nBytes; i++)
                 {
@@ -823,7 +753,7 @@ bool PImport::DoImport()
             {
                 PolyStringObject * ps = (PolyStringObject *)p;
                 /* The length is the number of characters. */
-                fscanf(f, "%lu", &nBytes);
+                fscanf(f, "%" POLYUFMT, &nBytes);
                 ch = getc(f); ASSERT(ch == '|');
                 ps->length = nBytes;
                 for (i = 0; i < nBytes; i++)
@@ -838,12 +768,14 @@ bool PImport::DoImport()
             }
 
         case 'C': /* Code segment. */
+        case 'D':
             {
+                bool oldForm = ch == 'C';
                 byte *u = (byte*)p;
                 POLYUNSIGNED length = p->Length();
                 /* Read the number of bytes of code and the number of words
                    for constants. */
-                fscanf(f, "%lu,%lu", &nWords, &nBytes);
+                fscanf(f, "%" POLYUFMT ",%" POLYUFMT, &nWords, &nBytes);
                 /* Read the code. */
                 ch = getc(f); ASSERT(ch == '|');
                 for (i = 0; i < nBytes; i++)
@@ -857,11 +789,14 @@ bool PImport::DoImport()
                 ASSERT(ch == '|');
                 /* Set the constant count. */
                 p->Set(length-1, PolyWord::FromUnsigned(nWords));
-                p->Set(length-1-nWords-1, PolyWord::FromUnsigned(0)); /* Profile count. */
-                p->Set(length-1-nWords-3, PolyWord::FromUnsigned(0)); /* Marker word. */
-                p->Set(length-1-nWords-2, PolyWord::FromUnsigned((length-1-nWords-2)*sizeof(PolyWord)));
-                /* Check - the code should end at the marker word. */
-                ASSERT(nBytes == ((length-1-nWords-3)*sizeof(PolyWord)));
+                if (oldForm)
+                {
+                    p->Set(length-1-nWords-1, PolyWord::FromUnsigned(0)); /* Profile count. */
+                    p->Set(length-1-nWords-3, PolyWord::FromUnsigned(0)); /* Marker word. */
+                    p->Set(length-1-nWords-2, PolyWord::FromUnsigned((length-1-nWords-2)*sizeof(PolyWord)));
+                    /* Check - the code should end at the marker word. */
+                    ASSERT(nBytes == ((length-1-nWords-3)*sizeof(PolyWord)));
+                }
                 /* Read in the constants. */
                 for (i = 0; i < nWords; i++)
                 {
@@ -878,9 +813,9 @@ bool PImport::DoImport()
                     while (ch != '\n')
                     {
                         ungetc(ch, f);
-                        unsigned long offset;
+                        POLYUNSIGNED offset;
                         int code;
-                        fscanf(f, "%lu,%d", &offset, &code);
+                        fscanf(f, "%" POLYUFMT ",%d", &offset, &code);
                         ch = getc(f);
                         ASSERT(ch == ',');
                         PolyWord constVal = TAGGED(0);
@@ -900,11 +835,11 @@ bool PImport::DoImport()
                 StackObject *s = (StackObject*)p;
                 POLYUNSIGNED n;
                 POLYUNSIGNED length = p->Length();
-                fscanf(f, "%lu", &nWords);
+                fscanf(f, "%" POLYUFMT, &nWords);
                 ch = getc(f); ASSERT(ch == '|');
 
                 /* Standard fields: size, pc, sp, hr. */
-                fscanf(f, "%lu", &s->p_space);
+                fscanf(f, "%" POLYUFMT, &s->p_space);
                 ch = getc(f); ASSERT(ch == ',');
                 if (! ReadValue(p, (PolyWord*)&s->p_pc - (PolyWord*)p))
                     return false;
@@ -916,7 +851,7 @@ bool PImport::DoImport()
                     return false;
 
                 /* Checked registers. */
-                fscanf(f, "%lu", &n);
+                fscanf(f, "%" POLYUFMT, &n);
                 s->p_nreg = n;
                 ch = getc(f); ASSERT(ch == '|');
                 for (i = 0; i < n; i++)
@@ -928,18 +863,18 @@ bool PImport::DoImport()
                            (ch == ' ' && i == n-1));
                 }
                 /* Unchecked registers. */
-                fscanf(f, "%lu", &n);
+                fscanf(f, "%" POLYUFMT, &n);
                 s->p_reg[i] = PolyWord::FromUnsigned(n);
                 ch = getc(f); ASSERT(ch == '|');
                 for (i = 0; i < n; i++)
                 {
                     POLYSIGNED n;
-                    fscanf(f, "%ld", &n);
+                    fscanf(f, "%" POLYSFMT, &n);
                     s->p_reg[s->p_nreg+i+1] = PolyWord::FromSigned(n);
                     ch = getc(f);
                 }
                 /* Stack values. */
-                fscanf(f, "%lu", &n);
+                fscanf(f, "%" POLYUFMT, &n);
                 ASSERT(n == length - (s->p_sp-(PolyWord*)p));
                 ch = getc(f); ASSERT(ch == '|');
                 for (i = 0; i < n; i++)
