@@ -57,7 +57,7 @@
 #define SAVE(x) taskData->saveVec.push(x)
 
 static const char *poly_runtime_system_copyright =
-"Copyright (c) 2002-10 CUTS, David C.J. Matthews and contributors.";
+"Copyright (c) 2002-12 CUTS, David C.J. Matthews and contributors.";
 
 // Property bits for functions.  For compiled functions these are
 // stored in the register mask word.  None of the architectures has
@@ -88,27 +88,48 @@ static POLYUNSIGNED rtsProperties(TaskData *taskData, int i)
         // This isn't quite true.  It is possible to clear the mutable flag on a mutable segment.
         return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_str_compare: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
-    case POLY_SYS_teststreq: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
-    case POLY_SYS_teststrneq: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_teststrgtr: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_teststrlss: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_teststrgeq: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_teststrleq: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_exception_trace: return 0;
     case POLY_SYS_give_ex_trace: return 0;
+    case POLY_SYS_exception_trace_fn: return 0;
+    case POLY_SYS_give_ex_trace_fn: return 0;
     case POLY_SYS_lockseg: return PROPWORD_NORAISE|PROPWORD_NODEREF;
     case POLY_SYS_emptystring: return 0; // Not a function
     case POLY_SYS_nullvector: return 0; // Not a function
     case POLY_SYS_network: return 0;
     case POLY_SYS_os_specific: return 0;
+    case POLY_SYS_eq_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_neq_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_geq_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_leq_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_gt_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_lt_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_io_dispatch: return 0;
     case POLY_SYS_signal_handler: return 0;
+    case POLY_SYS_atomic_reset: return PROPWORD_NORAISE;
     case POLY_SYS_atomic_incr: return PROPWORD_NORAISE;
     case POLY_SYS_atomic_decr: return PROPWORD_NORAISE;
     case POLY_SYS_thread_self: return PROPWORD_NORAISE|PROPWORD_NOUPDATE;
     case POLY_SYS_thread_dispatch: return 0;
+    case POLY_SYS_plus_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_minus_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_mul_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_div_longword: return PROPWORD_NOUPDATE|PROPWORD_NODEREF; // Can raise Divide
+    case POLY_SYS_mod_longword: return PROPWORD_NOUPDATE|PROPWORD_NODEREF; // Can raise Divide
+    case POLY_SYS_andb_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_orb_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_xorb_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_kill_self: return 0;
+    case POLY_SYS_shift_left_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_shift_right_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_shift_right_arith_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_profiler: return 0;
+    case POLY_SYS_longword_to_tagged: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_signed_to_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_unsigned_to_longword: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_full_gc: return PROPWORD_NORAISE; // Effectively has a side-effect
     case POLY_SYS_stack_trace: return 0;
     case POLY_SYS_timing_dispatch: return 0;
@@ -160,9 +181,11 @@ static POLYUNSIGNED rtsProperties(TaskData *taskData, int i)
     case POLY_SYS_io_operation: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_set_code_constant: return 0;
     case POLY_SYS_move_words: return PROPWORD_NORAISE;
+    case POLY_SYS_move_words_overlap: return PROPWORD_NORAISE;
     case POLY_SYS_shift_right_arith_word: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_int_to_word: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_move_bytes: return PROPWORD_NORAISE;
+    case POLY_SYS_move_bytes_overlap: return PROPWORD_NORAISE;
     case POLY_SYS_code_flags: return 0;
     case POLY_SYS_shrink_stack: return 0;
     case POLY_SYS_stderr: return 0; // Not a function
@@ -182,6 +205,10 @@ static POLYUNSIGNED rtsProperties(TaskData *taskData, int i)
     case POLY_SYS_int_leq: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_int_gtr: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_int_lss: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+        // Load_byte/word_immut are the same as load_byte/word except that they can only
+        // be applied to immutables so will always return the same result.
+    case POLY_SYS_load_byte_immut: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
+    case POLY_SYS_load_word_immut: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_mul_word: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_plus_word: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
     case POLY_SYS_minus_word: return PROPWORD_NORAISE|PROPWORD_NOUPDATE|PROPWORD_NODEREF;
@@ -204,6 +231,7 @@ static POLYUNSIGNED rtsProperties(TaskData *taskData, int i)
     }
 }
 
+#if (0)
 // Convert the statistics into ML data.  This is further unpicked within ML.
 static Handle unpackStats(TaskData *taskData, const polystatistics *stats)
 {
@@ -212,7 +240,7 @@ static Handle unpackStats(TaskData *taskData, const polystatistics *stats)
     for (unsigned i = 0; i < N_PS_COUNTERS; i++)
     {
         Handle mark = taskData->saveVec.mark();
-        Handle counterValue = Make_unsigned(taskData, stats->psCounters[i]);
+        Handle counterValue = Make_arbitrary_precision(taskData, stats->psCounters[i]);
         counts->WordP()->Set(i, counterValue->Word());
         taskData->saveVec.reset(mark);
     }
@@ -224,7 +252,7 @@ static Handle unpackStats(TaskData *taskData, const polystatistics *stats)
     for (unsigned j = 0; j < N_PS_SIZES; j++)
     {
         Handle mark = taskData->saveVec.mark();
-        Handle sizeValue = Make_unsigned(taskData, stats->psSizes[j]);
+        Handle sizeValue = Make_arbitrary_precision(taskData, stats->psSizes[j]);
         sizes->WordP()->Set(j, sizeValue->Word());
         taskData->saveVec.reset(mark);
     }
@@ -266,6 +294,7 @@ static Handle unpackStats(TaskData *taskData, const polystatistics *stats)
     resultVec->WordP()->Set(3, users->Word());
     return resultVec;
 }
+#endif
 
 Handle poly_dispatch_c(TaskData *taskData, Handle args, Handle code)
 {
@@ -350,21 +379,9 @@ Handle poly_dispatch_c(TaskData *taskData, Handle args, Handle code)
     case 24: // Return the name of the immediate parent stored in a child
         return ShowParent(taskData, args);
 
-    case 25: // Get the local statistics
-        {
-            polystatistics localStats;
-            if (! globalStats.getLocalsStatistics(&localStats))
-                raise_exception_string(taskData, EXC_Fail, "No statistics available");
-            return unpackStats(taskData, &localStats);
-        }
-
-    case 26: // Get remote statistics.  The argument is the process ID to get the statistics.
-        {
-            polystatistics localStats;
-            if (! globalStats.getRemoteStatistics(get_C_ulong(taskData, DEREFHANDLE(args)), &localStats))
-                raise_exception_string(taskData, EXC_Fail, "No statistics available");
-            return unpackStats(taskData, &localStats);
-        }
+    case 25: // Old statistics - now removed
+    case 26:
+        raise_exception_string(taskData, EXC_Fail, "No statistics available");
 
     case 27: // Get number of user statistics available
         return Make_arbitrary_precision(taskData, N_PS_USER);
@@ -374,11 +391,17 @@ Handle poly_dispatch_c(TaskData *taskData, Handle args, Handle code)
             unsigned index = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(0));
             if (index >= N_PS_USER)
                 raise_exception0(taskData, EXC_subscript);
-            int value = get_C_int(taskData, DEREFHANDLE(args)->Get(1));
+            POLYSIGNED value = get_C_long(taskData, DEREFHANDLE(args)->Get(1));
             globalStats.setUserCounter(index, value);
             Make_arbitrary_precision(taskData, 0);
         }
 
+    case 29: // Get local statistics.
+        return globalStats.getLocalStatistics(taskData);
+
+    case 30: // Get remote statistics.  The argument is the process ID to get the statistics.
+        return globalStats.getRemoteStatistics(taskData, get_C_ulong(taskData, DEREFHANDLE(args)));
+
     case 50: // GCD
         return gcd_arbitrary(taskData, SAVE(DEREFHANDLE(args)->Get(0)), SAVE(DEREFHANDLE(args)->Get(1)));
     case 51: // LCM
@@ -386,12 +409,11 @@ Handle poly_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
         // These next ones were originally in process_env and have now been moved here,
     case 100: /* Return the maximum word segment size. */
-            return Make_arbitrary_precision(taskData, MAX_OBJECT_SIZE);
+            return taskData->saveVec.push(TAGGED(MAX_OBJECT_SIZE));
     case 101: /* Return the maximum string size (in bytes).
                  It is the maximum number of bytes in a segment
                  less one word for the length field. */
-            return Make_arbitrary_precision(taskData,
-                (MAX_OBJECT_SIZE)*sizeof(PolyWord) - sizeof(PolyWord));
+            return taskData->saveVec.push(TAGGED((MAX_OBJECT_SIZE)*sizeof(PolyWord) - sizeof(PolyWord)));
     case 102: /* Test whether the supplied address is in the io area.
                  This was previously done by having get_flags return
                  256 but this was changed so that get_flags simply
@@ -418,7 +440,7 @@ Handle poly_dispatch_c(TaskData *taskData, Handle args, Handle code)
                     {
                         int regMask = machineDependent->GetIOFunctionRegisterMask(i);
                         POLYUNSIGNED props = rtsProperties(taskData, i);
-                        return Make_arbitrary_precision(taskData, regMask | props);
+                        return taskData->saveVec.push(TAGGED(regMask | props));
                     }
                 }
                 raise_exception_string(taskData, EXC_Fail, "Io pointer not found");
