@@ -110,7 +110,13 @@
 // "environ" is declared in the headers on some systems but not all.
 // Oddly, declaring it within process_env_dispatch_c causes problems
 // on mingw where "environ" is actually a function.
+#if __APPLE__
+// On Mac OS X there may be problems accessing environ directly.
+#include <crt_externs.h>
+#define environ (*_NSGetEnviron())
+#else
 extern char **environ;
+#endif
 
 /* Functions registered with atExit are added to this list. */
 static PolyWord at_exit_list = TAGGED(0);
@@ -120,6 +126,26 @@ static bool exiting = false;
 
 static PLock atExitLock; // Thread lock for above.
 
+#ifdef __CYGWIN__
+// Cygwin requires spawnvp to avoid the significant overhead of vfork
+// but it doesn't seem to be thread-safe.  Run it on the main thread
+// to be sure.
+class CygwinSpawnRequest: public MainThreadRequest
+{
+public:
+    CygwinSpawnRequest(char **argv): MainThreadRequest(MTP_CYGWINSPAWN), spawnArgv(argv) {}
+
+    virtual void Perform();
+    char **spawnArgv;
+    int pid;
+};
+
+void CygwinSpawnRequest::Perform()
+{
+    pid = spawnvp(_P_NOWAIT, "/bin/sh", spawnArgv);
+}
+
+#endif
 
 Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
 {
@@ -193,9 +219,9 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
             argv[2] = buff;
             argv[3] = NULL;
 #if (defined(__CYGWIN__))
-            // vfork adds a significant overhead.  spawnvp provides a wrapper
-            // for CreateProcess that avoids this.
-            int pid = spawnvp(_P_NOWAIT, "/bin/sh", argv);
+            CygwinSpawnRequest request(argv);
+            processes->MakeRootRequest(mdTaskData, &request);
+            int pid = request.pid;
             if (pid < 0)
                 raise_syscall(mdTaskData, "Function system failed", errno);
 #else
@@ -234,7 +260,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                             if (! fResult)
                                 raise_syscall(mdTaskData, "Function system failed", -(int)GetLastError());
                             CloseHandle((HANDLE)pid);
-                            return Make_unsigned(mdTaskData, result);
+                            return Make_arbitrary_precision(mdTaskData, result);
                         }
                     case WAIT_FAILED:
                         raise_syscall(mdTaskData, "Function system failed", -(int)GetLastError());
