@@ -95,6 +95,7 @@ public:
 
 NORETURNFN(extern Handle exitThread(TaskData *mdTaskData));
 
+Handle AtomicReset(TaskData *taskData, Handle mutexp);
 Handle AtomicIncrement(TaskData *taskData, Handle mutexp);
 Handle AtomicDecrement(TaskData *taskData, Handle mutexp);
 Handle ThreadSelf(TaskData *taskData);
@@ -117,6 +118,7 @@ extern enum _mainThreadPhase {
     MTP_LOADSTATE,
     MTP_PROFILING,
     MTP_SIGHANDLER,
+    MTP_CYGWINSPAWN,
     MTP_MAXENTRY
 } mainThreadPhase;
 
@@ -193,6 +195,8 @@ public:
     virtual ~ProcessExternal() {} // Defined to suppress a warning from GCC
 
     virtual TaskData *GetTaskDataForThread(void) = 0;
+    virtual TaskData *CreateNewTaskData(Handle threadId, Handle threadFunction,
+                           Handle args, PolyWord flags) = 0;
     // Request all ML threads to exit and set the result code.  Does not cause
     // the calling thread itself to exit since this may be called on the GUI thread.
     virtual void Exit(int n) = 0;
@@ -252,6 +256,7 @@ public:
 
 // Return the number of processors.  Used when configuring multi-threaded GC.
 extern unsigned NumberOfProcessors(void);
+extern unsigned NumberOfPhysicalProcessors(void);
 
 extern ProcessExternal *processes;
 
