@@ -3,6 +3,7 @@
 
     Copyright (c) 2000
         Cambridge University Technical Services Limited
+    Further development copyright (c) David C.J. Matthews 2011
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -32,7 +33,9 @@ typedef enum {
     kProfileOff = 0,
     kProfileTime,
     kProfileStoreAllocation,
-    kProfileEmulation
+    kProfileEmulation,
+    kProfileLiveData,
+    kProfileLiveMutables
 } ProfileMode;
 
 extern ProfileMode profileMode;
@@ -41,6 +44,8 @@ extern ProfileMode profileMode;
 
 extern void handleProfileTrap(TaskData *taskData, SIGNALCONTEXT *context);
 extern Handle profilerc(TaskData *taskData, Handle mode_handle);
-extern void add_count(TaskData *taskData, POLYCODEPTR pc, PolyWord *sp, int incr);
+extern void add_count(TaskData *taskData, POLYCODEPTR pc, PolyWord *sp, POLYUNSIGNED incr);
+extern void AddObjectProfile(PolyObject *obj);
+extern void printprofile();
 
 #endif /* _PROFILING_H_DEFINED */
