@@ -5,6 +5,8 @@
     Copyright (c) 2000, 2009
         Cambridge University Technical Services Limited
 
+    Further work copyright David C. J. Matthews 2012
+
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; either
@@ -275,6 +277,7 @@ Handle make_exn(TaskData *taskData, int id, Handle arg)
     case EXC_foreign: exName = "Foreign"; break;
     case EXC_Fail: exName = "Fail"; break;
     case EXC_thread: exName = "Thread"; break;
+    case EXC_extrace: exName = "ExTrace"; break;
     default: ASSERT(0); exName = "Unknown"; // Shouldn't happen.
     }
    
@@ -376,83 +379,58 @@ void raise_fail(TaskData *taskData, const char *errmsg)
     raise_exception_string(taskData, EXC_Fail, errmsg);
 }
 
-
-/******************************************************************************/
-/*                                                                            */
-/*      TRACE FUNCTIONS                                                       */
-/*                                                                            */
-/******************************************************************************/
-
-bool trace_allowed = false; // Allows ^C to abort a trace.
-
-/******************************************************************************/
-/*                                                                            */
-/*      give_stack_trace - utility function - doesn't allocate                */
-/*                                                                            */
-/******************************************************************************/
-void give_stack_trace(TaskData *taskData, PolyWord *sp, PolyWord *finish)
+// Build a list of the function names on the stack.
+Handle buildStackList(TaskData *taskData, PolyWord *startOfTrace, PolyWord *endOfTrace)
 {
-    /* Now search for the return addresses on the stack.
-       The values we find on the stack which are not PolyWord aligned may be
-       either return addresses or the addresses of handlers. */
-
-    trace_allowed = true; /* May be switch off by catchINT. */
-    
-    // The exception handler is used to suppress the addresses of
-    // handlers which would otherwise look like return addresses.
-    // Since we don't pass that in we may find it is actually out of
-    // date if we are producing a trace as a result of pressing ^C.
+    Handle saved = taskData->saveVec.mark();
+    Handle list = SAVE(ListNull);
     StackObject *stack = taskData->stack->stack();
-    PolyWord *exceptions = stack->p_hr;
-    PolyWord *endStack = taskData->stack->top;
-    
-#ifdef DEBUG    
-    printf("starting trace: sp = %p, finish = %p, end_of_stack = %p\n",
-        sp, finish, endStack);
-    fflush(stdout);
-#endif
-    
-    if (finish > endStack) finish = endStack;
-    
-    for(; trace_allowed && sp < finish-1; sp++)
+    PolyWord *endStack = taskData->stack->top - 1;
+    if (endOfTrace > endStack) endOfTrace = endStack;
+
+    for (PolyWord *sp = endOfTrace; sp >= startOfTrace; sp--)
     {
         PolyWord pc = *sp;
-        
-        /* If this is an exception handler do not treat it as return
-           address, just get the next handler */
-        if (sp == exceptions)
+        if (pc.IsCodePtr() && sp != stack->p_hr)
         {
-            /* Skip over the handlers until we find the next pointer up the
-                stack. */
-            while (sp < finish) {
-                exceptions = (*sp).AsStackAddr();
-                if (exceptions >= sp && exceptions <= endStack)
-                    break;
-                sp++;
-            }
-        }
-        else if (pc.IsCodePtr())
-        {
-           /* A code pointer will be either a return address or a pointer
-              to the constant section. (Or an exception handler?) */
-            // We used to have a check that this was not a constant area
-            // pointer but we don't have those any more.
-            
-            /* Initialise ptr to points at the end-of-code marker */
-            // This used to use the OBJ_CODEPTR_TO_CONSTS_PTR macro which
-            // simply looked for the end-of-code marker.
+            // A code pointer can be a return address or an exception
+            // handler but if we're producing an exception trace the
+            // only exception handler will be the one for exception
+            // trace itself.
             PolyObject *ptr = ObjCodePtrToPtr(pc.AsCodePtr());
             PolyWord *consts = ptr->ConstPtrForCode();
-            PolyWord p_name = consts[0]; /* Get procedure name */
-            
-            /* The name may be zero if it is anonymous */
-            if (p_name == TAGGED(0)) fputs("<anon>\n",stdout);
-            else {
-                print_string(p_name);
-                putc('\n',stdout);
-            }
+
+            // The name may be zero if it is anonymous.
+            // We have to be careful that a GC might move the code or the name.
+            // Stack areas are no longer in the ML heap so we don't need to worry
+            // about the stack pointer.
+            Handle functionName =
+                consts[0] == TAGGED(0) ? SAVE(C_string_to_Poly(taskData, "<anon>")) : SAVE(consts[0]);
+            Handle next  = alloc_and_save(taskData, sizeof(ML_Cons_Cell) / sizeof(PolyWord));
+
+            DEREFLISTHANDLE(next)->h = DEREFWORDHANDLE(functionName); 
+            DEREFLISTHANDLE(next)->t = DEREFLISTHANDLE(list);
+
+            taskData->saveVec.reset(saved);
+            list = SAVE(DEREFHANDLE(next));
         }
     }
+
+    return list;
+}
+
+void give_stack_trace(TaskData *taskData, PolyWord *sp, PolyWord *finish)
+{
+    Handle listHandle = buildStackList(taskData, sp, finish);
+    PolyWord list = listHandle->Word();
+
+    while (! (list.IsTagged()))
+    {
+        ML_Cons_Cell *p = (ML_Cons_Cell*)list.AsObjPtr();
+        print_string(p->h);
+        putc('\n',stdout);
+        list = p->t;
+    }
     fflush(stdout);
 }
 
@@ -470,12 +448,7 @@ static Handle stack_trace_c(TaskData *taskData)
     return SAVE(TAGGED(0));
 }
 
-/******************************************************************************/
-/*                                                                            */
-/*      ex_tracec - called from assembly code                                 */
-/*                                                                            */
-/******************************************************************************/
-/* CALL_IO2(ex_trace, REF, REF, NOIND) */
+// Legacy exception trace
 Handle ex_tracec(TaskData *taskData, Handle exnHandle, Handle handler_handle)
 {
     PolyWord *handler = DEREFWORD(handler_handle).AsStackAddr();
@@ -504,7 +477,7 @@ Handle ex_tracec(TaskData *taskData, Handle exnHandle, Handle handler_handle)
     give_stack_trace(taskData, stack->p_sp, handler);
     fputs("End of trace\n\n",stdout);
     fflush(stdout);
-    
+
     /* Set up the next handler so we don't come back here when we raise
        the exception again. */
     taskData->stack->stack()->p_hr = (PolyWord*)(handler->AsStackAddr());
@@ -516,7 +489,27 @@ Handle ex_tracec(TaskData *taskData, Handle exnHandle, Handle handler_handle)
     /*NOTREACHED*/
 }
 
-/* end of interrupt handling */
+// Current exception trace.  This creates a special exception packet and
+// raises it so that the ML code can print the trace.
+Handle exceptionToTraceException(TaskData *taskData, Handle exnHandle)
+{
+    StackObject *stack = taskData->stack->stack();
+    // p_hr points at a pair of values.  The first word will be the
+    // entry point to the handler i.e. to this code, the second word is
+    // the stack address of the handler to restore.
+    PolyWord *handler = stack->p_hr+1;
+    Handle listHandle = buildStackList(taskData, stack->p_sp, handler);
+    // Construct a pair of the trace list and the exception packet
+    Handle pair = alloc_and_save(taskData, 2);
+    pair->WordP()->Set(0, listHandle->Word());
+    pair->WordP()->Set(1, exnHandle->Word());
+    // Set up the next handler so we don't come back here when we raise
+    // the exception again. */
+    taskData->stack->stack()->p_hr = (PolyWord*)(handler->AsStackAddr());
+    // Raise this as an exception.  The handler will be able to
+    // print the trace and reraise the exception.
+    raise_exception(taskData, EXC_extrace, pair);
+}
 
 // Return the address of the iovec entry for a given index.
 Handle io_operation_c(TaskData *taskData, Handle entry)
@@ -673,7 +666,7 @@ static Handle testBytesEqual(TaskData *taskData, Handle len, Handle yOffset, Han
 static Handle vec_length_c(TaskData *taskData, Handle vector)    /* Length of a vector */
 {
     POLYUNSIGNED length = vector->WordP()->Length();
-    return Make_arbitrary_precision(taskData, length);
+    return taskData->saveVec.push(TAGGED(length));
 }
 
 static Handle load_byte_long_c(TaskData *taskData, Handle byte_no /* offset in BYTES */, Handle addr)
@@ -873,14 +866,183 @@ static Handle shift_right_word_c(TaskData *taskData, Handle y, Handle x)
 
 static Handle shift_right_arith_word_c(TaskData *taskData, Handle y, Handle x)
 {
-    POLYUNSIGNED wx = UNTAGGED_UNSIGNED(DEREFWORD(x)); /* Treat as a signed quantity. */
+    POLYSIGNED wx = UNTAGGED(DEREFWORD(x)); /* Treat as a signed quantity. */
     POLYUNSIGNED wy = UNTAGGED_UNSIGNED(DEREFWORD(y));
-    /* It is defined to return 0 or ~1 if the shift is greater than the
+    // This function in ML is defined to return 0 or ~1 if the shift is greater
+    // than the number of bits in the word.
+    // C does not actually define how signed values are shifted although most
+    // (all?) compilers seem to use arithmetic shifts here.
+    if (wy > sizeof(PolyWord)*8)
+        return taskData->saveVec.push(wx < 0 ? TAGGED(-1) : TAGGED(0));
+    return taskData->saveVec.push(TAGGED(wx >> wy));
+}
+
+// Large-word operations.  A large word is a 32/64-bit value in a byte segment.
+// These will normally be code-generated and in the assembly code.
+static Handle eqLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return taskData->saveVec.push(wx==wy ? TAGGED(1) : TAGGED(0));
+}
+
+static Handle neqLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return taskData->saveVec.push(wx!=wy ? TAGGED(1) : TAGGED(0));
+}
+
+static Handle geqLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return taskData->saveVec.push(wx>=wy ? TAGGED(1) : TAGGED(0));
+}
+
+static Handle leqLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return taskData->saveVec.push(wx<=wy ? TAGGED(1) : TAGGED(0));
+}
+
+static Handle gtLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return taskData->saveVec.push(wx>wy ? TAGGED(1) : TAGGED(0));
+}
+
+static Handle ltLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return taskData->saveVec.push(wx<wy ? TAGGED(1) : TAGGED(0));
+}
+
+static Handle makeLongWord(TaskData *taskData, POLYUNSIGNED w)
+{
+    Handle result = alloc_and_save(taskData, 1, F_BYTE_OBJ);
+    result->WordP()->Set(0, PolyWord::FromUnsigned(w));
+    return result;
+}
+
+static Handle plusLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return makeLongWord(taskData, wx + wy);
+}
+
+static Handle minusLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return makeLongWord(taskData, wx - wy);
+}
+
+static Handle mulLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return makeLongWord(taskData, wx * wy);
+}
+
+static Handle divLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    if (wy == 0) raise_exception0(taskData, EXC_divide);
+    return makeLongWord(taskData, wx / wy);
+}
+
+static Handle modLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    if (wy == 0) raise_exception0(taskData, EXC_divide);
+    return makeLongWord(taskData, wx % wy);
+}
+
+static Handle andbLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return makeLongWord(taskData, wx & wy);
+}
+
+static Handle orbLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return makeLongWord(taskData, wx | wy);
+}
+
+static Handle xorbLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = y->WordP()->Get(0).AsUnsigned();
+    return makeLongWord(taskData, wx ^ wy);
+}
+
+static Handle shiftLeftLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    // The amount to shift is always a Word.word value.
+    POLYUNSIGNED wy = UNTAGGED_UNSIGNED(DEREFWORD(y));
+    /* It is defined to return 0 if the shift is greater than the
+       number of bits in the PolyWord.  The shift instructions on many
+       architectures don't get that right. */
+    if (wy > sizeof(PolyWord)*8)
+        return makeLongWord(taskData, 0);
+    return makeLongWord(taskData, wx << wy);
+}
+
+static Handle shiftRightLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    POLYUNSIGNED wy = UNTAGGED_UNSIGNED(DEREFWORD(y));
+    /* It is defined to return 0 if the shift is greater than the
        number of bits in the word.  The shift instructions on many
        architectures don't get that right. */
     if (wy > sizeof(PolyWord)*8)
-        return taskData->saveVec.push(wx < 0 ? TAGGED(-1) : TAGGED(0));
-    return taskData->saveVec.push(TAGGED(wx>>wy));
+        return makeLongWord(taskData, 0);
+    return makeLongWord(taskData, wx >> wy);
+}
+
+static Handle shiftRightArithLongWord(TaskData *taskData, Handle y, Handle x)
+{
+    POLYSIGNED wx = x->WordP()->Get(0).AsSigned();
+    POLYUNSIGNED wy = UNTAGGED_UNSIGNED(DEREFWORD(y));
+    if (wy > sizeof(PolyWord)*8)
+        return makeLongWord(taskData, wx < 0 ? -1 : 0);
+    // Strictly speaking, C does not define that this uses an arithmetic shift
+    // so we really ought to set the high-order bits explicitly.
+    return makeLongWord(taskData, (POLYUNSIGNED)(wx >> wy));
+}
+
+// Extract the first word and return it as a tagged value.  This loses the top-bit
+static Handle longWordToTagged(TaskData *taskData, Handle x)
+{
+    POLYUNSIGNED wx = x->WordP()->Get(0).AsUnsigned();
+    return taskData->saveVec.push(TAGGED(wx));
+}
+
+// Shift the tagged value to remove the tag and put it into the first word.
+// The original sign bit is copied in the shift.
+static Handle signedToLongWord(TaskData *taskData, Handle x)
+{
+    POLYSIGNED wx = x->Word().UnTagged();
+    return makeLongWord(taskData, (POLYUNSIGNED)wx);
+}
+
+// As with the above except the value is treated as an unsigned
+// value and the top bit is zero.
+static Handle unsignedToLongWord(TaskData *taskData, Handle x)
+{
+    POLYUNSIGNED wx = x->Word().UnTaggedUnsigned();
+    return makeLongWord(taskData, wx);
 }
 
 static Handle set_code_constant(TaskData *taskData, Handle data, Handle constant, Handle offseth, Handle base)
@@ -1019,14 +1181,6 @@ Handle EnterPolyCode(TaskData *taskData)
                 machineDependent->CallIO2(taskData, compareStrings);
                 break;
 
-            case POLY_SYS_teststreq:
-                machineDependent->CallIO2(taskData, &testStringEqual);
-                break;
-
-            case POLY_SYS_teststrneq:
-                machineDependent->CallIO2(taskData, &testStringNotEqual);
-                break;
-
             case POLY_SYS_teststrgtr:
                 machineDependent->CallIO2(taskData, &testStringGreater);
                 break;
@@ -1044,7 +1198,13 @@ Handle EnterPolyCode(TaskData *taskData)
                 break;
 
             case POLY_SYS_exception_trace: // Special case.
-                machineDependent->SetExceptionTrace(taskData);
+                // This is the legacy version. 
+                machineDependent->SetExceptionTrace(taskData, true);
+                break;
+
+            case POLY_SYS_exception_trace_fn: // Special case.
+                // This is the current version. 
+                machineDependent->SetExceptionTrace(taskData, false);
                 break;
 
     //        case POLY_SYS_lockseg: machineDependent->CallIO1(taskData, &locksegc); break;
@@ -1199,6 +1359,10 @@ Handle EnterPolyCode(TaskData *taskData)
                 machineDependent->CallIO1(taskData, &io_operation_c);
                 break;
 
+            case POLY_SYS_atomic_reset:
+                machineDependent->CallIO1(taskData, &AtomicReset);
+                break;
+
             case POLY_SYS_atomic_incr:
                 machineDependent->CallIO1(taskData, &AtomicIncrement);
                 break;
@@ -1278,10 +1442,12 @@ Handle EnterPolyCode(TaskData *taskData)
                 break;
 
             case POLY_SYS_load_byte:
+            case POLY_SYS_load_byte_immut:
                 machineDependent->CallIO2(taskData, &load_byte_long_c);
                 break;
 
             case POLY_SYS_load_word:
+            case POLY_SYS_load_word_immut:
                 machineDependent->CallIO2(taskData, &load_word_long_c);
                 break;
 
@@ -1369,10 +1535,12 @@ Handle EnterPolyCode(TaskData *taskData)
                 break;
 
             case POLY_SYS_move_bytes:
+            case POLY_SYS_move_bytes_overlap:
                 machineDependent->CallIO5(taskData, &move_bytes_long_c);
                 break;
 
             case POLY_SYS_move_words:
+            case POLY_SYS_move_words_overlap:
                 machineDependent->CallIO5(taskData, &move_words_long_c);
                 break;
 
@@ -1432,12 +1600,96 @@ Handle EnterPolyCode(TaskData *taskData)
                 machineDependent->CallIO0(taskData, exitThread);
                 break;
 
+            case POLY_SYS_eq_longword:
+                machineDependent->CallIO2(taskData, &eqLongWord);
+                break;
+
+            case POLY_SYS_neq_longword:
+                machineDependent->CallIO2(taskData, &neqLongWord);
+                break;
+
+            case POLY_SYS_geq_longword:
+                machineDependent->CallIO2(taskData, &geqLongWord);
+                break;
+
+            case POLY_SYS_leq_longword:
+                machineDependent->CallIO2(taskData, &leqLongWord);
+                break;
+
+            case POLY_SYS_gt_longword:
+                machineDependent->CallIO2(taskData, &gtLongWord);
+                break;
+
+            case POLY_SYS_lt_longword:
+                machineDependent->CallIO2(taskData, &ltLongWord);
+                break;
+
+            case POLY_SYS_plus_longword:
+                machineDependent->CallIO2(taskData, &plusLongWord);
+                break;
+
+            case POLY_SYS_minus_longword:
+                machineDependent->CallIO2(taskData, &minusLongWord);
+                break;
+
+            case POLY_SYS_mul_longword:
+                machineDependent->CallIO2(taskData, &mulLongWord);
+                break;
+
+            case POLY_SYS_div_longword:
+                machineDependent->CallIO2(taskData, &divLongWord);
+                break;
+
+            case POLY_SYS_mod_longword:
+                machineDependent->CallIO2(taskData, &modLongWord);
+                break;
+
+            case POLY_SYS_andb_longword:
+                machineDependent->CallIO2(taskData, &andbLongWord);
+                break;
+
+            case POLY_SYS_orb_longword:
+                machineDependent->CallIO2(taskData, &orbLongWord);
+                break;
+
+            case POLY_SYS_xorb_longword:
+                machineDependent->CallIO2(taskData, &xorbLongWord);
+                break;
+
+            case POLY_SYS_shift_left_longword:
+                machineDependent->CallIO2(taskData, &shiftLeftLongWord);
+                break;
+
+            case POLY_SYS_shift_right_longword:
+                machineDependent->CallIO2(taskData, &shiftRightLongWord);
+                break;
+
+            case POLY_SYS_shift_right_arith_longword:
+                machineDependent->CallIO2(taskData, &shiftRightArithLongWord);
+                break;
+
+            case POLY_SYS_longword_to_tagged:
+                machineDependent->CallIO1(taskData, &longWordToTagged);
+                break;
+
+            case POLY_SYS_signed_to_longword:
+                machineDependent->CallIO1(taskData, &signedToLongWord);
+                break;
+
+            case POLY_SYS_unsigned_to_longword:
+                machineDependent->CallIO1(taskData, &unsignedToLongWord);
+                break;
+
             // This is called from assembly code and doesn't actually have an entry in the
             // io vector.
             case POLY_SYS_give_ex_trace:
                 machineDependent->CallIO2(taskData, ex_tracec);
                 break;
 
+            case POLY_SYS_give_ex_trace_fn:
+                machineDependent->CallIO1(taskData, exceptionToTraceException);
+                break;
+
             default:
                 Crash("Unknown io operation %d\n", ioFunction);
             }
