@@ -390,8 +390,8 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 4: /* Return a constant. */
         {
-            unsigned i = get_C_long(taskData, DEREFWORDHANDLE(args));
-            if (i < 0 || i >= sizeof(unixConstVec)/sizeof(unixConstVec[0]))
+            unsigned i = get_C_unsigned(taskData, DEREFWORDHANDLE(args));
+            if (i >= sizeof(unixConstVec)/sizeof(unixConstVec[0]))
                 raise_syscall(taskData, "Invalid index", 0);
             return Make_arbitrary_precision(taskData, unixConstVec[i]);
         }
@@ -429,28 +429,28 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
     case 9: /* get real user id */
         {
             uid_t uid = getuid();
-            if (uid < 0) raise_syscall(taskData, "getuid failed", errno);
+            // This is defined always to succeed
             return Make_arbitrary_precision(taskData, uid);
         }
 
     case 10: /* get effective user id */
         {
             uid_t uid = geteuid();
-            if (uid < 0) raise_syscall(taskData, "geteuid failed", errno);
+            // This is defined always to succeed
             return Make_arbitrary_precision(taskData, uid);
         }
 
     case 11: /* get real group id */
         {
             gid_t gid = getgid();
-            if (gid < 0) raise_syscall(taskData, "getgid failed", errno);
+            // This is defined always to succeed
             return Make_arbitrary_precision(taskData, gid);
         }
 
     case 12: /* get effective group id */
         {
             gid_t gid = getegid();
-            if (gid < 0) raise_syscall(taskData, "getegid failed", errno);
+            // This is defined always to succeed
             return Make_arbitrary_precision(taskData, gid);
         }
 
@@ -682,12 +682,16 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
     case 29: /* uname */ return getUname(taskData);
 
     case 30: /* Get controlling terminal. */
+#ifdef HAVE_CTERMID
         {
             char *term = ctermid(0);
             /* Can this generate an error? */
             if (term == 0) raise_syscall(taskData, "ctermid failed", errno);
             return SAVE(C_string_to_Poly(taskData, term));
         }
+#else
+       raise_syscall(taskData, "ctermid is not implemented", 0);
+#endif
 
     case 31: /* Get terminal name for file descriptor. */
         {
@@ -1168,8 +1172,12 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
                process until it has. */
             PIOSTRUCT strm = get_stream(args->WordP());
             if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
+#ifdef HAVE_TCDRAIN
             if (tcdrain(strm->device.ioDesc) < 0)
                 raise_syscall(taskData, "tcdrain failed", errno);
+#else
+            raise_syscall(taskData, "tcdrain is not implemented", 0);
+#endif
             return Make_arbitrary_precision(taskData, 0);
         }
     
@@ -1375,11 +1383,12 @@ static Handle getStatInfo(TaskData *taskData, struct stat *buf)
     uidHandle = Make_arbitrary_precision(taskData, buf->st_uid);
     gidHandle = Make_arbitrary_precision(taskData, buf->st_gid);
     sizeHandle = Make_arbitrary_precision(taskData, buf->st_size);
-    /* TODO: The only really standard time fields give the seconds part.  There
-       are various extensions which would give us microseconds or nanoseconds. */
-    atimeHandle = Make_arb_from_pair_scaled(taskData, buf->st_atime, 0, 1000000);
-    mtimeHandle = Make_arb_from_pair_scaled(taskData, buf->st_mtime, 0, 1000000);
-    ctimeHandle = Make_arb_from_pair_scaled(taskData, buf->st_ctime, 0, 1000000);
+    atimeHandle = Make_arb_from_pair_scaled(taskData, STAT_SECS(buf,a),
+                                            STAT_USECS(buf,a), 1000000);
+    mtimeHandle = Make_arb_from_pair_scaled(taskData, STAT_SECS(buf,m),
+                                            STAT_USECS(buf,m), 1000000);
+    ctimeHandle = Make_arb_from_pair_scaled(taskData, STAT_SECS(buf,c),
+                                            STAT_USECS(buf,c), 1000000);
     result = ALLOC(11);
     DEREFHANDLE(result)->Set(0, DEREFWORDHANDLE(modeHandle));
     DEREFHANDLE(result)->Set(1, DEREFWORDHANDLE(kindHandle));
